@@ -1,7 +1,7 @@
 import numpy as np
-from pyg_timeseries._math import stdev_calculation, skew_calculation
+from pyg_timeseries._math import stdev_calculation, skew_calculation, _w
 from pyg_timeseries._decorators import compiled, first_, _data_state
-from pyg_base import pd2np, Dict, is_num, loop_all, loop, clock
+from pyg_base import pd2np, Dict, is_num, loop_all, loop, clock, is_pd, df_reindex
 
 __all__ = ['ffill', 'bfill', 'fnna', 'na2v', 'v2na', 'diff', 'shift', 'ratio', 'rolling_mean', 'rolling_sum', 'rolling_rms', 'rolling_std', 'rolling_skew', 
            'diff_', 'shift_', 'ratio_', 'rolling_mean_', 'rolling_sum_', 'rolling_rms_', 'rolling_std_', 'rolling_skew_']
@@ -115,24 +115,26 @@ def _ffill(a, n = 0, prev = None, i = None):
         return _ffill2d(a, n, prev , i)
 
 
-@loop_all
+@loop(dict, list)
 @pd2np
-@compiled
 def _na2v(a, new = 0.0):
     res = a.copy()
-    for j in range(a.shape[0]):
-        if np.isnan(a[j]):
-            res[j] = new
+    if len(a.shape) == 1:
+        res[np.isnan(res)] = new
+    else:
+        for j in range(a.shape[0]):
+            res[j, np.isnan(res[j])] = new
     return res
 
-@loop_all
+@loop(dict, list)
 @pd2np
-@compiled
 def _v2na(a, old = 0.0, new = np.nan):
-    res = a.copy()
-    for j in range(a.shape[0]):
-        if a[j] == old:
-            res[j] = new
+    res = a.copy().astype(float)
+    if len(a.shape) == 1:
+        res[res == old] = new
+    else:
+        for j in range(a.shape[0]):
+            res[j, res[j] == old] = new
     return res
 
 
@@ -171,6 +173,33 @@ def _bfill(a, limit = -1):
     return res
 
 
+@loop_all
+@pd2np
+@compiled
+def _ewfill(a, fwd_n, bwd_n = np.nan, prev = np.nan, nxt = np.nan, decay_target = 0.):
+    res = a.copy()
+    prev = decay_target if np.isnan(prev) else prev
+    nxt = decay_target if np.isnan(nxt) else nxt
+    fw = _w(fwd_n)
+    bw = fw if np.isnan(bwd_n) else _w(bwd_n)
+    f = b = 1.
+    for j in range(a.shape[0]): ## forward sweep
+        if np.isnan(a[j]):
+            f *= fw
+            res[j] = 0.5 * (prev * f  + (1 - f) * decay_target)
+        else:
+            f = 1.
+            prev = res[j]    
+    for j in range(a.shape[0] - 1,  -1, -1):
+        if np.isnan(a[j]):
+            b *= bw
+            res[j] += 0.5*(nxt * b  + (1 - b) * decay_target)
+        else:
+            b = 1.
+            nxt = res[j]
+    return res
+
+
 ###############
 ##
 ## bottleneck
@@ -252,6 +281,40 @@ def _diff1(a, vec, time, i = 0, t = np.nan):
     return res, vec, i, t
 
 
+@loop_all
+@pd2np
+@compiled
+def _buffer(a, band, unit = 0.0, pos = 0):
+    """
+    >>> a = pd.Series(cumsum(np.random.normal(0,1,1000)), drange(-999))
+    >>> ts, pos = _buffer(a, 1, 3, 0)
+    >>> df_concat([a,ts])[dt(-100):].plot()
+    """
+    res = np.full(a.shape, np.nan)
+    b = 0.0
+    if np.isnan(pos):
+        pos = 0.0
+    for i in range(a.shape[0]):
+        if not np.isnan(a[i]):
+            if not np.isnan(band[i]): ## we forward fill band
+                b = band[i]
+            if pos < a[i] - b:
+                pos = a[i] - b
+                if unit > 0:
+                    pos = np.round(pos / unit) * unit
+                    if pos < a[i] - b and pos + unit < a[i] + b:
+                        pos += unit
+            elif pos > a[i] + b:
+                pos = a[i] + b
+                if unit > 0:
+                    pos = np.round(pos / unit) * unit
+                    if pos > a[i] + b and pos - unit > a[i] - b:
+                        pos -= unit
+            res[i] = pos
+    return res, pos
+
+
+
 @loop_all
 def _tdiff(a, n, vec, i, time = None, t = None):
     time = clock(a, time, t)
@@ -466,6 +529,7 @@ def fnna(a, n=1, axis = 0):
     """
     return _fnna(a, n, axis = axis)
 
+
 def fnnz(a, n=1, axis = 0):
     """
     returns the index in a of the nth first non-nan and non-zero.
@@ -487,6 +551,35 @@ def fnnz(a, n=1, axis = 0):
     return _fnnz(a, n, axis = axis)
 
 
+def ewfill(a, fwd_n, bwd_n = None, axis = 0, decay_target = 0.0):
+    """
+    returns a double-direction exponentially-weighted filled result.
+    
+    The idea is that we decay value to long term target (say 0). 
+    The weight we get from previous observation decays exponentiall (using fwd_n)
+    The weight we get from next observation decays exponentiall backwards (using bwd_n)
+    The remaining weight is assigned to long term decay_target
+    
+    :Parameters:
+    ------------
+    a : array/timeseries
+        array/timeseries
+    fwd_n: int/float, optional
+        speed of fwd filling decay weight
+    bwd_n: int/float, optional
+        speed of bwd_n filling decay weight
+    state: dict, optional
+        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
+
+    :Example:
+    ---------
+    >>> a = np.array([np.nan,np.nan,1,np.nan,np.nan,2,np.nan,np.nan,np.nan])
+    >>> fnna(a, n=-2)
+    """
+    bwd_n = bwd_n or fwd_n
+    return _ewfill(a, fwd_n = fwd_n, bwd_n = bwd_n, decay_target = decay_target)
+
+
 
 def bfill(a, n = -1, axis = 0):
     """
@@ -569,6 +662,8 @@ def v2na(a, old = 0.0, new = np.nan):
     array/timeseries
 
     """
+    if is_num(a):
+        return new if a == old else a
     return _v2na(a, old = old, new = new)
 
 def na2v(a, new = 0.0):
@@ -593,6 +688,8 @@ def na2v(a, new = 0.0):
     array/timeseries
 
     """
+    if is_num(a):
+        return new if np.isnan(a) else a
     return _na2v(a, new)
     
 
@@ -625,6 +722,8 @@ def init2v(a, n = 0, new = np.nan):
     return _init2v(a, n, new)
 
 
+    
+
 def diff(a, n=1, time = None, axis = 0, data = None, state = None):
     """
     equivalent to a.diff(n) in pandas if there are no nans. If there are, we SKIP nans rather than propagate them.
@@ -656,6 +755,8 @@ def diff(a, n=1, time = None, axis = 0, data = None, state = None):
     """
     if n == 0:
         return a - a
+    if is_num(state) and n == 1:
+        state = Dict(vec = np.array([state]), i = 0)
     state = state or Dict(vec = None, i = 0, t = np.nan)
     state.vec = _vec(a,state.vec, n, axis=axis)
     return first_(_tdiff1(a, time = time, axis = axis, **state) if n == 1 
@@ -676,6 +777,50 @@ def diff_(a, n=1, time = None, axis = 0, data = None, instate = None):
                        else _tdiff(a, n=n, time = time, axis = axis, **state))
 
 diff_.output = ['data', 'state']
+
+def buffer_(a, band, unit = 0.0, data = None, instate = None):
+    if is_num(instate):
+        instate = Dict(pos = instate)
+    elif instate is None:
+        instate = Dict(pos = 0.0)
+    if is_num(band):
+        band = np.full(a.shape, band)
+    return _data_state(['data', 'pos'], _buffer(a = a, band = band, unit = unit, **instate))
+        
+buffer_.output = ['data', 'state']
+
+def buffer(a, band, unit = 0.0, data = None, state = None):
+    """
+    buffer performs two functions:
+        - ensures the result is stated in 'units' so if unit == 1, output is integers
+        - slows down the trading to ensure we are always within [a-band, a+band] but does not change values while we are within the band
+    
+    :Parameters:
+    ------------
+    a : array/timeseries
+        array/timeseries
+    band: float
+        band size
+    unit: float
+        the resulting timeseries will always be whole number of units
+    data: None.
+        unused at the moment. 
+    state: dict, float, optional
+        state parameters of last position, used to instantiate the internal calculations, based on history prior to 'a' provided. 
+        please provide the variable "pos" if a dict, or just a float, indicating previous position.
+    """
+    if is_num(state):
+        state = Dict(pos = state)
+    elif state is None:
+        state = Dict(pos = 0.0)
+    if is_num(band):
+        band = np.full(a.shape, band)
+    if is_pd(band) and is_pd(a):
+        band = df_reindex(band, a, method = 'ffill')
+    return first_(_buffer(a = a, band = band, unit = unit, **state))
+    
+        
+
         
 def shift(a, n=1, axis = 0, data = None, state = None):
     """
@@ -718,7 +863,7 @@ def shift(a, n=1, axis = 0, data = None, state = None):
     >>> new_ts = shift(new, **old_ts)
     >>> assert eq(new_ts, shift(a).iloc[3:])
     """
-    if n == 0:
+    if n == 0 or is_num(a):
         return a
     state = state or Dict(vec = None, i = 0,)
     state.vec = _vec(a, state.vec, n, axis=axis)
@@ -729,7 +874,7 @@ def shift_(a, n=1, axis = 0, instate = None):
     Equivalent to shift(a,n) but returns the full state. See shift for full details
   
     """
-    if n == 0:
+    if n == 0 or is_num(a):
         return Dict(data = a, state = instate)
     state = instate or Dict(vec = None, i = 0,)
     state.vec = _vec(a, state.vec, n, axis=axis)
@@ -760,14 +905,19 @@ def ratio(a, n=1, time = None, data = None, state = None, axis = 0):
     >>> assert eq(ratio(a), pd.Series([np.nan, 2, 1.5, 4/3,1.25], drange(-4)))
     >>> assert eq(ratio(a,2), pd.Series([np.nan, np.nan, 3, 2, 5/3], drange(-4)))
     """
+    if is_num(a):
+        return 1
     state = state or Dict(vec = None, i = 0, t = np.nan)
     state.vec = _vec(a, state.vec, n, axis=axis)
     return first_(_tratio(a, n, axis = axis, time = time, **state))
 
-def ratio_(a, n=1, time = None, axis = 0, data = None, instate = None):
+def ratio_(a, n=1, time = None, axis = 0, data = None, instate = None):        
     state = instate or Dict(vec = None, i = 0, t = np.nan) 
     state.vec = _vec(a, state.vec, n, axis=axis)
-    return _data_state(['data', 'vec', 'i', 't'], _tratio(a, n, time = time, **state))
+    if is_num(a):
+        return Dict(data = 1, state = state)
+    else:
+        return _data_state(['data', 'vec', 'i', 't'], _tratio(a, n, time = time, **state))
 
 ratio_.output = ['data', 'state']
 
